use crate::server::ServerType;
use clap::{Args, Parser, Subcommand};
use std::path::PathBuf;

#[derive(Parser)]
#[command(author, version, about, long_about = None)]
pub struct Cli {
    #[command(subcommand)]
    pub command: Commands,
    /// Directory where configs are stored, and where the server will run
    #[arg(
        long,
        value_name = "CONFIG_DIR",
        default_value = ".",
        env = "ALEX_CONFIG_DIR",
        global = true
    )]
    pub config: PathBuf,
    /// Directory where world files will be saved
    #[arg(
        long,
        value_name = "WORLD_DIR",
        default_value = "../worlds",
        env = "ALEX_WORLD_DIR",
        global = true
    )]
    pub world: PathBuf,
    /// Directory where backups will be stored
    #[arg(
        long,
        value_name = "BACKUP_DIR",
        default_value = "../backups",
        env = "ALEX_BACKUP_DIR",
        global = true
    )]
    pub backup: PathBuf,

    /// How many backups to keep
    #[arg(
        short = 'n',
        long,
        default_value_t = 7,
        env = "ALEX_MAX_BACKUPS",
        global = true
    )]
    pub max_backups: u64,
}

#[derive(Subcommand)]
pub enum Commands {
    /// Run the server
    Run(RunArgs),
    /// Create a new backup of the server. This command should only be used when the server is not
    /// running.
    Backup(BackupArgs),
}

#[derive(Args)]
pub struct RunArgs {
    /// Type of server
    pub type_: ServerType,
    /// Version string for the server, e.g. 1.19.4-545
    #[arg(env = "ALEX_SERVER_VERSION")]
    pub server_version: String,

    /// Server jar to execute
    #[arg(
        long,
        value_name = "JAR_PATH",
        default_value = "server.jar",
        env = "ALEX_JAR"
    )]
    pub jar: PathBuf,

    /// Java command to run the server jar with
    #[arg(long, value_name = "JAVA_CMD", default_value_t = String::from("java"), env = "ALEX_JAVA")]
    pub java: String,

    /// XMS value in megabytes for the server instance
    #[arg(long, default_value_t = 1024, env = "ALEX_XMS")]
    pub xms: u64,
    /// XMX value in megabytes for the server instance
    #[arg(long, default_value_t = 2048, env = "ALEX_XMX")]
    pub xmx: u64,

    /// How frequently to perform a backup, in minutes; 0 to disable.
    #[arg(short = 't', long, default_value_t = 0, env = "ALEX_FREQUENCY")]
    pub frequency: u64,

    /// Don't actually run the server, but simply output the server configuration that would have
    /// been ran
    #[arg(short, long, default_value_t = false)]
    pub dry: bool,
}

#[derive(Args)]
pub struct BackupArgs {
    /// Type of server
    pub type_: ServerType,
    /// Version string for the server, e.g. 1.19.4-545
    #[arg(env = "ALEX_SERVER_VERSION")]
    pub server_version: String,
}
